#include "include/common.h"

#define	CLIENT_KEEP_OR_BREAK	{if ( self.keep_alive == TRUE ) {	\
					client_session_init( &self );	\
					goto request_continue;		\
				}else break;}

void *client( void *data )
{
	extern struct config_t	config;

	int			client_fd = *((int *) data);
	struct thread_t		self;

	struct pollfd		pollfd;
	int			poll_res;
	int			read_res;

	int			db_res;
	int			result;

	int			http_res;

	memset( &self, 0, sizeof(self) );
	self.init = FALSE;
	self.status = THREAD_INIT;
	self.time_init = time( &self.time_init );

	if ( client_fd <= 0 )
		pthread_exit(0);

	self.client_fd = client_fd;

	self.request.buf_size = config.buf_size_recv;
	self.request.buf = malloc( self.request.buf_size );

	self.response.header_buf = malloc( config.buf_size_header );

	self.response.buf_size = config.buf_size_send;
	self.response.buf = malloc( self.response.buf_size );

	self.tid = pthread_self();

	pollfd.fd = client_fd;
	pollfd.events = POLLIN | POLLPRI | POLLERR | POLLHUP;

	monitor_thread_push( &self );
	self.init = TRUE;

	while ( 1 )
	{
request_continue:
		self.status = THREAD_WAIT;
		poll_res = poll( &pollfd, 1, config.http_timeout*1000 );

		if ( poll_res <= 0 )
			break;

		if ( self.request.buf_size-self.request.read_len <= 1024*8 )
		{
			self.request.buf = realloc( self.request.buf, self.request.buf_size + LIMIT_BUF_INCREMENT );
			self.request.buf_size += LIMIT_BUF_INCREMENT;
		}

		read_res = read( self.client_fd, &(self.request.buf[self.request.read_len]), self.request.buf_size-self.request.read_len );

		if ( read_res <= 0 )
			client_safe_exit( &self );

		self.request.read_len += read_res;

		//printf("%s\n", self.request.buf);

		if ( self.request.header_len <= 0 )
		{
			self.request.buf[self.request.read_len] = '\0';
			request_header_parse( &self.request );
		}

		if ( self.request.header_len > 0 ) 
		{
			switch ( request_end( &self.request ) )
			{
				case SQLITE_OK:
				break;
				case SQLITE_EMPTY:
					goto request_continue;
				break;
				case SQLITE_ERROR:
					if ( self.request.header.method == REQUEST_METHOD_GET )
						break;
					response_error( &self, RESPONSE_ERR_BAD_REQUEST, NULL );
					goto client_error;
				break;
			}

			self.request.buf[self.request.read_len] = '\0';

			self.keep_alive = request_keep_alive( &self.request );

			self.status = THREAD_WORK;

			if ( self.request.header.method == REQUEST_METHOD_POST )
			{
				self.time_last = time( &self.time_last );

				if ( client_auth_proc( &self ) != SQLITE_OK )
				{
					//auth error
					if ( response_error( &self, RESPONSE_ERR_UNAUTHORIZED, NULL ) != TRUE )
						break;
					CLIENT_KEEP_OR_BREAK
				}

				db_res = client_open_db( &self );

				if ( self.super_user == 1 )
				{
					if ( db_res == SQLITE_NOTFOUND || db_res == SQLITE_OK )
					{
						result = admin_process( &self );
						if ( result == SQLITE_OK )
							CLIENT_KEEP_OR_BREAK
						else if ( result == SQLITE_ERROR_ADMIN )
							goto normal_operation_start;
						else if ( result == SQLITE_ERROR_SEND )
							break;
					}
				}

normal_operation_start:
				if ( db_res == SQLITE_OK )
				{
					//db process
					result = client_process( &self, self.request.data );
					if ( result == SQLITE_OK )
						//process
						CLIENT_KEEP_OR_BREAK
					else if ( result == SQLITE_ERROR_SEND )
						break;
				}
				else
				{
					if ( response_error( &self, RESPONSE_ERR_DB_FOUND, NULL ) != TRUE )
						break;
					CLIENT_KEEP_OR_BREAK
					//server error
				}
			}
			else
			{
				if ( config.http_use == TRUE )
				{
					http_res = http_proc( &self );
					if ( http_res == SQLITE_OK )
					{
						CLIENT_KEEP_OR_BREAK
					}
					else
						break;
				}
				else
				{
					if ( response_error( &self, RESPONSE_ERR_BAD_REQUEST, NULL ) != TRUE )
						break;
					CLIENT_KEEP_OR_BREAK
					//bad request
				}
			}
			CLIENT_KEEP_OR_BREAK
		}
	}

client_error:

#ifdef __DEBUG
	printf("***** Client thread terminated.\n");
#endif

	client_safe_exit( &self );
	
	pthread_exit( 0 );
}

int client_auth_proc( struct thread_t *self )
{
	extern struct db_t		db_auth;
	extern pthread_mutex_t		db_auth_mutex;

	char				*cookie;
	char				*cookie_value;
	short				cookie_auth;
	char				*auth;
	int				auth_len;
	char				*base64;

	char				iddata[1024];
	int				iddata_len;

	char				*sql;
	char				*id;
	char				*pass;

	char				*res_super;

	struct timeval			now;
	char				tmp[64];
	int				tmp_len;

	cookie_auth = FALSE;

	if ( request_get_char( &cookie, &self->request, "COOKIE" ) > 0 && strlen(self->session_key) > 0 )
	{
		cookie_value = strstr( cookie, "SQLSESSION=" );
		if ( cookie_value != NULL )
		{
			cookie_value += 11;
			if ( strncmp( cookie_value, self->session_key, MIN( 32, MIN( strlen(cookie_value), strlen(self->session_key) ) ) ) == 0 )
			{
				return SQLITE_OK;
			}
		}
	}

	auth_len = request_get_char( &auth, &self->request, "AUTHORIZATION" );
	if ( auth == NULL || auth_len < 0 )
		return SQLITE_AUTH;

	base64 = strchr( auth, ' ' );
	if ( base64 == NULL )
		return SQLITE_AUTH;

	iddata_len = base64_decode( iddata, ++base64, strlen(base64) );
	if ( iddata_len <= 0 )
		return SQLITE_AUTH;
	iddata[iddata_len] = '\0';

	id = iddata;
	pass = strchr( iddata, ':' );
	if ( pass == NULL )
		return SQLITE_AUTH;
	pass[0] = '\0';
	pass++;

	sql = db_printf( CATALOG_GET_USER_SQL, id, pass, self->request.header.db_name );
	pthread_mutex_lock( &db_auth_mutex );
	db_get_table( &db_auth, sql );
	if ( db_table_has_rows( &db_auth ) != SQLITE_OK )
	{
		db_free_table( &db_auth );
		pthread_mutex_unlock( &db_auth_mutex );
		db_free( sql );
		return SQLITE_AUTH;
	}

	res_super = db_table_result( &db_auth, 0, 0 );

	db_free_table( &db_auth );
	pthread_mutex_unlock( &db_auth_mutex );
	db_free( sql );

	gettimeofday( &now, NULL );
	tmp_len = snprintf( tmp, 64, "%ld%ld", now.tv_usec, now.tv_sec );
	for ( ; tmp_len--; )
		tmp[tmp_len] += ( 17 + tmp[tmp_len]%(tmp_len+1) );

	strncpy( self->session_key, tmp, 32 );
	strncpy( self->id, id, 64 );
	self->response.set_cookie = 1;
	self->super_user = (memcmp( res_super, "1", 1 ) == 0)?1:0;

	return SQLITE_OK;
}

int client_open_db( struct thread_t *self )
{
	int	open_res;
	
	if ( strlen( self->request.header.db_name ) <= 0 )
		return SQLITE_NOTFOUND;

	if ( strncasecmp( self->db.db_name, self->request.header.db_name, self->request.header.db_name_len ) == 0 )
	{
		return SQLITE_OK;
	}

	open_res = db_open( &self->db, self->request.header.db_name );

	if ( open_res == SQLITE_OK )
	{
		memcpy( self->db.db_name, self->request.header.db_name, self->request.header.db_name_len );
		return SQLITE_OK;
	}
	return open_res;
}

int client_process( struct thread_t *self, char *query )
{
	int		result;
	int		change;
	long long int	rowid;
	int		rows;
	int		columns;
	char		*column_name;
	int		column_name_len;
	int		i, p;
	char		*data;
	int		data_len;

	result = db_get_table( &self->db, query );

	if ( result != SQLITE_OK )
		return response_error( self, RESPONSE_ERR_INTERNAL, self->db.errmsg );

	change = db_changes( &self->db );
	rowid = db_last_rowid( &self->db );
	rows = db_table_num_rows( &self->db );

	self->response.offset = sprintf( self->response.buf, XML_RESPONSE_OPEN, change, rowid, rows );

	//query success
	if ( db_table_has_rows( &self->db ) == SQLITE_OK )
	{
		columns = db_table_columns( &self->db );
		//columns open
		self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_COLUMNS_OPEN );
		for ( i=0; i < columns; i++ )
		{
			self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_COLUMN_OPEN );
			column_name_len = db_table_column_name( &column_name, &self->db, i );
			self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), column_name, column_name_len );
			self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_COLUMN_CLOSE );
		}
		self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_COLUMNS_CLOSE );

		for ( i=0; i < rows; i++ )
		{
			self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_ROW_OPEN );
			for ( p=0; p < columns; p++ )
			{
				self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_DATA_OPEN );
				data = db_table_result( &self->db, i, p );
				data_len = strlen( data );
				self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), data, data_len );
				self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_DATA_CLOSE );
			}
			self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_ROW_CLOSE );
		}
	}

	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_CLOSE );
	db_free_table( &self->db );

	self->response.status = 200;
	self->response.status_msg = "OK";

	return response( self );
}

void client_session_init( struct thread_t *self )
{
	self->keep_alive = FALSE;

	self->request.read_len = self->request.header_len = 0;
	self->request.data = NULL;
	self->request.data_len = self->request.header.node_idx = self->request.header.method = self->request.header.db_name_len = 0;
	self->request.header.db_name = NULL;

	self->response.offset = 0;
	self->response.header_send = self->response.set_cookie = FALSE;
	self->response.status = 0;
	self->response.status_msg = self->response.errmsg = NULL;
}

void client_safe_exit( struct thread_t *self )
{
	if ( self->request.buf != NULL )
		free( self->request.buf );
	if ( self->response.header_buf != NULL )
		free( self->response.header_buf );
	if ( self->response.buf != NULL )
		free( self->response.buf );
	close( self->client_fd );

	if ( self->db.db != NULL )
	{
		if ( self->db.auth == TRUE && self->db.transaction == TRUE )
		{
			if ( self->db.auto_commit == TRUE )
				db_commit( &self->db );
			else
				db_rollback( &self->db );
		}
		db_close( &self->db );
	}

	if ( self->init == TRUE )
		monitor_thread_delete( self );

	pthread_exit(0);
}

