#include "include/common.h"

static short	response_errstatus[] = {
	400, 401, 500, 406, 406, 406, 406, 406, 406, 404
};

static char	*response_errmsg[] = {
	"Bad Request",			"Bad Request",
	"Unauthorized",			"Authorization Required",
	"Internal Server Error",	"Internal Server Error",
	"Cannot create database",	"Not Acceptable",
	"Database not found",		"Not Acceptable",
	"Cannot drop database",		"Not Acceptable",
	"Cannot create user",		"Not Acceptable",
	"User not found",		"Not Acceptable",
	"Cannot drop user",		"Not Acceptable",
	"Resource not found",		"Not Found"
};

static char	*response_infomsg[] = {
	"Database created",
	"Database droped",
	"User created",
	"User droped",
	"User alterd"
};

int response_error( struct thread_t *self, int status, char *errmsg )
{
	self->response.status = response_errstatus[status];
	self->response.status_msg = response_errmsg[(status*2)+1];
	if ( errmsg == NULL )
		self->response.errmsg = response_errmsg[status*2];
	else
		self->response.errmsg = errmsg;

	self->response.offset = sprintf( self->response.buf, XML_ERROR_OPEN );
	self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), self->response.errmsg, strlen(self->response.errmsg) );
	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_ERROR_CLOSE );

	//printf("%s\n", self->response.buf);

	return response( self );
}

int response_info( struct thread_t *self, int status, char *infomsg )
{
	self->response.status = 202;
	self->response.status_msg = "Accepted";
	if ( infomsg == NULL )
		self->response.errmsg = response_infomsg[status];
	else
		self->response.errmsg = infomsg;

	self->response.offset = sprintf( self->response.buf, XML_INFO_OPEN );
	self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), self->response.errmsg, strlen(self->response.errmsg) );
	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_INFO_CLOSE );

	//printf("%s\n", self->response.buf);

	return response( self );
}

int response( struct thread_t *self )
{
	extern struct config_t	config;
	int			len;
	int			send_res;
	char			date_822[48];

	if ( self->response.status == 304 )
	{
		http_time2asctime( date_822, self->response.stat.st_mtime );
		len = snprintf( self->response.header_buf, 1024*4, "HTTP/1.1 %d %s\nServer: %s\nLast-Modified: %s\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, date_822 );
	}
	else if ( self->response.file > 0 )
	{
		http_time2asctime( date_822, self->response.stat.st_mtime );
		len = snprintf( self->response.header_buf, 1024*4, "HTTP/1.1 %d %s\nServer: %s\nContent-length: %d\nLast-Modified: %s\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, self->response.stat.st_size, date_822 );
	}
	else
	{
		len = snprintf( self->response.header_buf, 1024*4, "HTTP/1.1 %d %s\nServer: %s\nContent-type: text/xml\nContent-length: %d\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, self->response.offset );
	}

	if ( self->response.status == 401 )
		len += snprintf( &(self->response.header_buf[len]), (1024*4)-len, "WWW-Authenticate: Basic realm=SQLiteDBMS\n" );

	if ( self->response.set_cookie )
		len += snprintf( &(self->response.header_buf[len]), (1024*4)-len, "Set-Cookie: SQLSESSION=%s; expire=0; path=/\n", self->session_key );

	if ( self->keep_alive == TRUE )
		len += snprintf( &(self->response.header_buf[len]), (1024*4)-len, "Connection: Keep-Alive\nKeep-Alive: timeout=%d\n\n", config.http_timeout );
	else
		len += snprintf( &(self->response.header_buf[len]), (1024*4)-len, "Connection: Close\n\n" );

#ifdef __DEBUG
	printf("%s\n", self->response.header_buf);
	printf("%s\n", self->response.buf);
#endif

	if ( net_send( self->client_fd, self->response.header_buf, len ) != SQLITE_OK )
		return SQLITE_ERROR_SEND;

	if ( self->request.header.method == REQUEST_METHOD_HEAD || self->response.status == 304 )
		return SQLITE_OK;

	if ( self->response.file > 0 )
	{
		lseek( self->response.file, 0, self->response.file_offset );
		return net_sendfile( self->client_fd, self->response.file, self->response.file_offset, self->response.stat.st_size );
	}
	else
		return net_send( self->client_fd, self->response.buf, self->response.offset );
}

int response_xml_printf( char *dst, const char *src, const int src_len )
{
	int		offset;
	int		i;

	offset = i = 0;
	for ( ; i < src_len; i++ )
	{
		switch ( src[i] )
		{
			case '&':
				dst[offset++] = '&';
				dst[offset++] = 'a';
				dst[offset++] = 'm';
				dst[offset++] = 'p';
				dst[offset++] = ';';
			break;
			case '"':
				dst[offset++] = '&';
				dst[offset++] = 'q';
				dst[offset++] = 'u';
				dst[offset++] = 'o';
				dst[offset++] = 't';
				dst[offset++] = ';';
			break;
			case '\'':
				dst[offset++] = '&';
				dst[offset++] = '#';
				dst[offset++] = '0';
				dst[offset++] = '3';
				dst[offset++] = '9';
				dst[offset++] = ';';
			break;
			case '<':
				dst[offset++] = '&';
				dst[offset++] = 'l';
				dst[offset++] = 't';
				dst[offset++] = ';';
			break;
			case '>':
				dst[offset++] = '&';
				dst[offset++] = 'g';
				dst[offset++] = 't';
				dst[offset++] = ';';
			break;
			default:
				dst[offset++] = src[i];
			break;
		}
	}
	return offset;
}
