typedef union {
	struct sockaddr		sa;
	struct sockaddr_in	sa_in;
} sock_t;

struct config_node_t {
	char			*buf;
	char			*field;
	int			field_len;
	char			*value;
	int			value_len;
};

struct config_t {
	char			*filepath;
	int			server_port;
	int			max_session;

	int			poll_timer;

	char			*server_user;

	int			buf_size_header;
	int			buf_size_recv;
	int			buf_size_send;

	char			*db_path;
	char			*rep_path;

	short			http_use;
	char			*http_root;
	int			http_timeout;

	char			*log_path;
};

struct request_node_t {
	char			*field;
	int			field_len;
	char			*value;
	int			value_len;
};

struct request_header_t {
	int			method;
	char			*db_name;
	int			db_name_len;

	struct request_node_t	node[47];				//request field, HTTP/1.1 RFC
	int			node_idx;
};

struct request_t {
	char			*buf;
	int			buf_size;
	int			read_len;

	struct request_header_t	header;
	int			header_len;

	char			*data;
	int			data_len;
};

struct response_t {
	char			*header_buf;

	char			*buf;
	int			buf_size;
	int			offset;

	short			header_send;
	short			set_cookie;

	short			status;
	char			*status_msg;
	char			*errmsg;

	struct stat		stat;
	int			file;
	off_t			file_offset;
};

struct db_t {
	sqlite3			*db;					//sqlite3 db pointer
	char			db_name[LIMIT_MAX_DBNAME];		//sqlite3 db name, db filename exclude ".db"
	char			db_filename[LIMIT_MAX_DBFILENAME];	//sqlite3 db full file name
	short			opened;
	char			*errmsg;				//sqlite3 lib error msg

	sqlite3_stmt		*stmt;

	char			**table;
	int			table_res;
	int			table_nrow;
	int			table_ncol;

	short			transaction;
	short			auto_commit;

	short			auth;
};

struct thread_t {
	/*
	pthread_t		thread;
	pthread_mutex_t		thread_mutex;
	pthread_cond_t		thread_cond;
	*/
	short			init;
	short			status;
	time_t			time_init;
	time_t			time_last;

	int			client_fd;

	struct request_t	request;
	struct response_t	response;

	char			session_key[32];
	char			id[64];
	short			super_user;

	short			keep_alive;

	struct db_t		db;

	/** thread stat */
	pthread_t		tid;
	struct thread_t		*next;
};

struct db_user_t {
	short			super_user;
	char			*db_name;
	char			*username;
	char			*password;
};

struct stat_t {
	struct thread_t		*first;
	pthread_mutex_t		lock;
	int			count;
};

